﻿// Copyright 2018 Consorsbank
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

using Grpc.Core;
using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace TestClient
{
    // tag::AsyncObserver[]
    public abstract class AsyncObserver<TClient, TRequest, TReply> where TClient : ClientBase<TClient>
    {
        // Define delegates
        public delegate void ResultListener(TReply reply); // <1>
        public delegate AsyncUnaryCall<TReply> Call(TRequest request, CancellationToken token); // <2>

        private readonly TClient _client;
        private readonly Call _caller;

        public AsyncObserver(TClient client, Call caller)
        {
            this._client = client;
            this._caller = caller;
        }

        public async Task<TReply> Process(TRequest request, ResultListener listener, 
            CancellationToken token = default(global::System.Threading.CancellationToken))
        {
            try
            {
                using (AsyncUnaryCall<TReply> call = _caller(request, token)) // <3>
                {
                    var responseStream = call.ResponseAsync; // <4>
                    await responseStream; // <5>
                    if (!responseStream.IsCanceled)
                    {
                        listener?.Invoke(responseStream.Result); // <6>
                    }
                    return responseStream.Result;
                }
            }
            catch (RpcException e)
            {
                if (e.StatusCode != StatusCode.Cancelled)
                {
                    Console.WriteLine(e); // <7>
                    throw;
                } else
                {
                    return default(TReply);
                }
            }
            catch (Exception e)
            {
                Console.WriteLine(e); // <8>
                throw;
            }

        }
    }
    // end::AsyncObserver[]

}
